# R class. "Handling DNA sequences in R: Biostrings package"
# Alexey Pindyurin
# 2015-02-03


# Empty workspace
rm(list=ls(all=TRUE));
ls();

# Set working directory
setwd('YOUR WORKING DIRECTORY!');
list.files(getwd());

# See what is in the workspace
ls();

#=============================================================================

# "Biostrings" package from Bioconductor:
# http://www.bioconductor.org/packages/release/bioc/html/Biostrings.html

# Essential (run once)!
source("http://bioconductor.org/biocLite.R");
# Essential (run once)!
biocLite("Biostrings");



library(Biostrings);

# Get information about the package
help(pac=Biostrings);

# the XString class used to represent a single string and 
# the XStringViews class used to represent a set of views on the same XString object

# The XString is in fact a virtual class and therefore cannot be instanciated. 
# Only subclasses (or subtypes) BString, DNAString, RNAString and AAString can.


# The BString class is a general container for storing a big string (a long 
# sequence of characters) and for making its manipulation easy and efficient

a <- BString("I am a BString object");
a;

length(a);

b <- a[length(a):1];


# The DNAString, RNAString and AAString classes are similar containers but with 
# the more biology-oriented purpose of storing a DNA sequence (DNAString), 
# an RNA sequence (RNAString), or a sequence of amino acids (AAString). 

# A DNAString object:

DNA <- DNAString("GGACCTCGCAACCTTGG------NNNNNNNNNNGGTCCAAGGTCCATCGNN-TGCGT");
DNA;

length(DNA);

RNA <- RNAString("GAUCCCUUAAGUCG");
RNA;

Protein <- AAString("RNDCQEGHILKMFP");
Protein;

#=============================================================================

GENETIC_CODE;

AMINO_ACID_CODE;

IUPAC_CODE_MAP;

#=============================================================================

# Sequence reversing and complementing

DNA;

reverse(DNA);

complement(DNA);

reverseComplement(DNA);


#=============================================================================

# Transcription & translation

cDNA <- DNAString("ATGAAATTGGTGGTTACCCTCATGTAA");

translate(cDNA);

translate(cDNA[c(1:3, 10:length(cDNA))]);


#=============================================================================

# Replacing letters in a sequence at some specified locations

DNA;

DNA2 <- replaceLetterAt(DNA, at=2, letter="A");
DNA2;

#=============================================================================

# Masking a sequence by motif

?maskMotif

DNA;

DNA.masked <- maskMotif(DNA, "GGACC");
DNA.masked;

DNA.masked <- maskMotif(DNA, "N");
DNA.masked;

DNA.masked2 <- maskMotif(DNA.masked, "-");
DNA.masked2;

gaps(DNA.masked2);

str(DNA.masked2);

DNA.masked3 <- as(DNA.masked2, "XStringViews");
DNA.masked3;

DNA.masked4 <- as(gaps(DNA.masked2), "XStringViews");
DNA.masked4;

DNA.masked;
DNA.unmasked <- unmasked(DNA.masked);
DNA.unmasked;

#-----------------------------------------------------------------------------

# Masking a sequence by position

DNA;
mask(DNA, start=2, end=15);

#=============================================================================

# String searching
# matchPattern(pattern, subject, algorithm="auto", max.mismatch=0, with.indels=FALSE, fixed=TRUE)
# fixed: If FALSE then IUPAC extended letters are interpreted as ambiguities 


DNA;

matchPattern(pattern = "GGACC", subject = DNA, algorithm="auto", max.mismatch=0, with.indels=FALSE, fixed=TRUE);

matchPattern("GGACC", DNA);

matchPattern("GGACC", DNA, fixed=FALSE);

matchPattern("GGACC", DNA.masked);

matchPattern("GGWCC", DNA, fixed=FALSE);

matchPattern("GGWCC", DNA, max.mismatch=1, fixed=FALSE);


motif <- DNAString("GGACC");
reverseComplement(motif);

matchPattern(reverseComplement(motif), DNA);


#=============================================================================

# Read FASTA formatted file

pNDamMyc <- readDNAStringSet(file="pNDamMyc.txt", format="fasta", nrec=-1L, skip=0L, use.names=TRUE);

class(pNDamMyc);
str(pNDamMyc);

pNDamMyc.seq <- DNAString(pNDamMyc[[1]])
pNDamMyc.seq;

# EcoRI restriction site(s)
matchPattern("GAATTC", pNDamMyc.seq);

# DpnI restriction site(s)
matchPattern("GATC", pNDamMyc.seq);


#=============================================================================

# Essential (run once)!
source("http://bioconductor.org/biocLite.R");
# Essential (run once)!
biocLite("BSgenome.Dmelanogaster.UCSC.dm3");


library(BSgenome.Dmelanogaster.UCSC.dm3);

Dmelanogaster;

str(Dmelanogaster);

length(Dmelanogaster);

length(seqnames(Dmelanogaster));

length(mseqnames(Dmelanogaster));

seqnames(Dmelanogaster);

mseqnames(Dmelanogaster);

#-----------------------------------------------------------------------------
# Getting a table with Drosophila chromosome lengths

chr.lengths <- rep(NA, length(seqnames(Dmelanogaster)));

for (i in 1:length(seqnames(Dmelanogaster)))
{
chr.lengths[i] <- length(DNAString(Dmelanogaster[[i]]))
}

chr.lengths;

Chr.lengths <- data.frame(seqnames(Dmelanogaster), chr.lengths);
Chr.lengths;

#-----------------------------------------------------------------------------

Dmelanogaster$chr3L

Dmelanogaster[[1]]

# Get complete sequence of the chromosome of interest
chrX.sequence <- DNAString(Dmelanogaster$chrX);
chr2L.sequence <- DNAString(Dmelanogaster$chr2L);
chr2R.sequence <- DNAString(Dmelanogaster$chr2R);
chr3L.sequence <- DNAString(Dmelanogaster$chr3L);
chr3R.sequence <- DNAString(Dmelanogaster$chr3R);
chr4.sequence <- DNAString(Dmelanogaster$chr4);
chrU.sequence <- DNAString(Dmelanogaster$chrU);


# Get a region of interest
ROI <- chr3L.sequence[18401188:18401206];
ROI;

ROI <- subseq(x=chr3L.sequence, start=18401188, end=18401206);
ROI;

ROI <- subseq(x=chr3L.sequence, start=18401188, width=19);
ROI;

ROI <- subseq(x=chr3L.sequence, start=18401188);
ROI;

ROI <- subseq(x=chr3L.sequence, end=18401188);
ROI;

ROI <- subseq(x=chr3L.sequence, end=18401206, width=19);
ROI;

ROI[2];

ROI[2:5];


#=============================================================================

# Function to calculate the frequency of letters in a biological sequence 

?alphabetFrequency

alphabetFrequency(x = chr3L.sequence, baseOnly=FALSE);

alphabetFrequency(x = chr3L.sequence, baseOnly=TRUE);

hasOnlyBaseLetters(x = chr3L.sequence);

uniqueLetters(x = chr3L.sequence);

dinucleotideFrequency(x = chr3L.sequence);

oligonucleotideFrequency(x = chr3L.sequence, width=4);

#-----------------------------------------------------------------------------

# Finding all DpnI sites in the Drosophila genome

startTime = Sys.time()

for (i in 1:length(seqnames(Dmelanogaster))){
   chr.i <- order(seqnames(Dmelanogaster), na.last=T, decreasing=F)[i];
   GATC.temp <- matchPattern("GATC", DNAString(Dmelanogaster[[chr.i]]), max.mismatch=0, fixed=T);
   GATC <- as.data.frame(matrix(data=NA, nrow=1+length(GATC.temp), ncol=4, byrow=F, dimnames=NULL));
   names(GATC) <- c("ID", "chr", "start", "end");

   GATC$chr <- sub("chr", "", seqnames(Dmelanogaster)[chr.i]);

   GATC$start[1] <- 1;
   GATC$start[2:(1+length(GATC.temp))] <- GATC.temp@ranges@start;

   GATC$end[1:(length(GATC.temp))] <- 3+GATC.temp@ranges@start;
   GATC$end[1+length(GATC.temp)] <- length(DNAString(Dmelanogaster[[chr.i]]));

   rm(GATC.temp);

   if (i == 1) GATCs <- GATC;
   if (i > 1)  GATCs <- rbind(GATCs, GATC);

   rm(chr.i);
   rm(GATC)
}
rm(i);


GATCs$ID <- paste("GATC.", seq(1, nrow(GATCs)), sep="");

print(Sys.time()-startTime)


dim(GATCs);

head(GATCs);

tail(GATCs);

GATCs$width <- GATCs$end-GATCs$start+1;

hist(GATCs$width);

min(GATCs$width);

which(GATCs$width==8);

GATCs[which(GATCs$width==8)[1],];

subseq(x=chr2L.sequence, start=GATCs$start[which(GATCs$width==8)[1]], end=GATCs$end[which(GATCs$width==8)[1]]);


max(GATCs$width);

which(GATCs$width==121514);

GATCs[which(GATCs$width==121514)[1],];

subseq(x=chrU.sequence, start=GATCs$start[which(GATCs$width==121514)[1]], end=GATCs$end[which(GATCs$width==121514)[1]]);


range(GATCs$width);

hist(GATCs$width, breaks=10000, xlim=c(0,5000));

hist(GATCs$width, breaks=10000, xlim=c(0,200));


#=============================================================================
